"""
Wrapper for tqdm progressbar with default settings.

Note about smoothing:
    Smoothing: 0=compute average speed, 1=compute current speed
    (math: EMAcurrent = (1 - smoothing) * EMAprevious + smoothing * valuecurrent)
    Default is 0.3
"""

from __future__ import annotations

from timeit import default_timer
from typing import Optional

from attrs import define, field
from tqdm import tqdm

from packg.strings.formatters import dict_to_str_comma_equals

TQDM_WID = 100


class tqdm_max_ncols(tqdm):
    """
    Wrapper for tqdm progressbar with a maximum width. Default max_ncols=90.

    This allows resizing a window later on without breaking the progressbar output.
    Usually tqdm would resize dynamically, but this doesn't always work (e.g. when using screen).
    """

    def __init__(self, *args, max_ncols: Optional[int] = TQDM_WID, **kwargs):
        super().__init__(*args, **kwargs)
        if self.disable:
            # pbar is disabled, no need to modify ncols
            return
        if max_ncols is not None and self.ncols is not None:
            self.ncols = min(self.ncols, max_ncols)


@define
class SimplePbar:
    """
    Usage example:
        from packg.tqdmext import SimplePbar
        sbar = SimplePbar(1000, len(mylist))
        for item in mylist:
            # do something
            sbar.update()
    """

    interval: int | None = None
    total: int | None = None
    print_fn: callable = print
    print_end: str = "\n"
    time_fmt: str = "m"
    t1: float = field(init=False)
    time_fmt_div: int = field(init=False)
    current: int = field(init=False)
    len_total_fmt: int = field(init=False)

    def __attrs_post_init__(self):
        self.reset()

    def reset(self):
        self.t1 = default_timer()
        self.current = 0
        self.len_total_fmt = 1
        if self.total is not None:
            self.len_total_fmt = len(str(self.total))

        if self.interval is None:
            if self.total is not None:
                self.interval = int(max(1, self.total // 10))
            else:
                self.interval = 1
        if self.time_fmt == "s":
            self.time_fmt_div = 1
        elif self.time_fmt == "m":
            self.time_fmt_div = 60
        elif self.time_fmt == "h":
            self.time_fmt_div = 3600
        elif self.time_fmt == "d":
            self.time_fmt_div = 86400
        else:
            raise ValueError(f"Invalid time_fmt {self.time_fmt}")

    def update(self, n: int = 1, status_counts: dict[str, int] | None = None):
        self.current += n
        if (self.current - 1) % self.interval != 0:
            return
        delta_t = default_timer() - self.t1
        delta_n = max(self.current, 1)
        current_fmt = f"{self.current:{self.len_total_fmt}d}"
        outputs = [f"{current_fmt}"]
        if self.total is not None:
            remaining_t = delta_t / delta_n * (self.total - self.current)
            outputs.append(
                f"/{self.total} Remaining {remaining_t / self.time_fmt_div:.2f}{self.time_fmt}"
            )
        outputs.append(f" Elapsed {delta_t / self.time_fmt_div:.2f}{self.time_fmt}")
        if status_counts is not None:
            outputs.append(f" {dict_to_str_comma_equals(dict(status_counts))}")
        output = " ".join(outputs)
        self.print_fn(output, end=self.print_end)



"""
The logger provided here is a link to the loguru.logger but sets default level to ERROR.
To activate logging, use the configure_logger function:

    >>> configure_logger(level=logging.DEBUG)

Other contents of this file: Utils for the standard library logging package,
Utils for the loguru package.
https://loguru.readthedocs.io/en/stable/resources/recipes.html#changing-the-level-of-an-existing-handler
"""

from __future__ import annotations

import itertools
import logging
import os
import sys
from copy import deepcopy
from logging import getLevelName
from typing import Union

from deprecated import deprecated
from loguru import logger as loguru_logger
from pathspec import PathSpec

from typedparser import VerboseQuietArgs

LevelType = Union[str, int]  # either "DEBUG" or 10
DEFAULT_LOGURU_FORMAT = (
    "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level: <8}</level> | "
    "<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - "
    "<level>{message}</level>"
)
SHORTER_FORMAT = (
    "<green>{time:YYYYMMDD HH:mm:ss}</green> <level>{level: <4.4}</level> "
    "<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> "
    "<level>{message}</level>"
)
SHORTEST_FORMAT = (
    "<green>{time:YYYYMMDD HH:mm:ss}</green> <level>{level: <4.4}</level> "
    "<level>{message}</level>"
)
BRIGHTBG_FORMAT = (
    "<green>{time:YYYYMMDD HH:mm:ss}</green> <level>{level: <4.4}</level> "
    "<blue>{name}</blue>:<blue>{function}</blue>:<blue>{line}</blue> "
    "<level>{message}</level>"
)
TIMELESS_FORMAT = "<level>{level: <4.4}</level> <level>{message}</level>"

SPINNER_STR = "|/-\\"
SPINNER_CYCLE = itertools.cycle("|/-\\")


def get_stdlib_logging_formatter():
    return logging.Formatter(
        "%(asctime)s %(levelname)-4s %(filename)s:%(lineno)d %(message)s",
        datefmt="%Y%m%d %H:%M:%S",
    )


global_logger_config = None


def configure_logger(
    level: LevelType = "DEBUG",
    sink=sys.stderr,
    format=SHORTEST_FORMAT,  # noqa # pylint: disable=redefined-builtin
    colorize=True,
    add_sinks: Union[list[any], list[dict[str, any]], None] = None,
    kwargs_handler: Union[dict[str, any], None] = None,
    **kwargs: any,
) -> dict[str, any]:
    """
    Configure the loguru logger. For more complex usages, use logger.configure() directly.

    Args:
        level: minimum level to log
        sink: where to write the logs to
        format: message formatting
        colorize: add color codes to the output
        add_sinks: list of other sinks to add. str sinks will write to file.
            use a dictionary to change parameters format, colorize, level separately from main sink.
        kwargs_handler: additional parameters to pass to each handler
        **kwargs: additional parameters to pass to logger.configure()

    References:
        https://loguru.readthedocs.io/en/stable/api/logger.html
        from loguru import _colorizer  # color code reference

    Returns:
        Configuration passed to logger.configure()

    """
    add_sinks = add_sinks if add_sinks is not None else []
    kwargs_handler = kwargs_handler if kwargs_handler is not None else {}
    sinks = [sink] + add_sinks
    handlers = []
    level_str = get_level_as_str(level)
    for lsink in sinks:
        if isinstance(lsink, dict):
            params = lsink
            if "format" not in params:
                params["format"] = format
            if "colorize" not in params:
                params["colorize"] = colorize
            if "level" not in params:
                params["level"] = level_str
        else:
            params = {
                "sink": lsink,
                "format": format,
                "colorize": colorize,
                "level": level_str,
            }
        assert "sink" not in kwargs_handler, f"sink is a reserved key, got {params}"
        params.update(deepcopy(kwargs_handler))
        if isinstance(lsink, str):
            # automatically disable color codes for files
            params["colorize"] = False
        handlers.append(params)

    # fix bug when combining colorama and loguru in conemu console on windows
    if (
        colorize
        and os.name == "nt"
        and os.environ.get("CONEMUANSI") == "ON"
        and sink in (sys.stderr, sys.stdout)
    ):
        print("\r", end="")  # the invisible print here magically fixes the color

    logger_config = {"handlers": handlers, **kwargs}
    loguru_logger.configure(**logger_config)
    global global_logger_config
    global_logger_config = logger_config
    return logger_config


def reroute_logger(new_sink=sys.stderr, logger_config=None, handler_num: int = 0) -> None:
    """
    Reroute a sink from one target to another. Useful for proper logging inside
    a tqdm progressbar without having to recreate the entire logger.

    Args:
        new_sink: new target
        logger_config: config created by configure_logger() function above.
            If None, use the global config.
        handler_num: index of the handler of which to change the sink

    Usage:
        logger_config = configure_logger(...)
        pbar = tqdm(...)
        reroute_logger(lambda msg: pbar.write(msg, end=""), logger_config)
        # use tqdm progressbar and logger together here
        pbar.close()
        reroute_logger(sys.stderr, logger_config)

    Notes:
        The config is modified inplace so does not need to be returned here.
        It cannot be deepcopy-ed here because sink is not pickle-able.
    """
    if logger_config is None:
        logger_config = global_logger_config
    logger_config["handlers"][handler_num]["sink"] = new_sink
    loguru_logger.configure(**logger_config)
    return logger_config


def get_level_as_str(level: LevelType):
    if isinstance(level, str):
        return level.upper()
    if isinstance(level, int):
        return getLevelName(level).upper()
    raise TypeError(f"Level must be str or int, not {type(level)}")


def get_level_as_int(level: LevelType):
    if isinstance(level, int):
        return level
    level_int = int(getLevelName(level.upper()))
    return level_int


def get_logger_level_from_args(args: VerboseQuietArgs) -> str:
    if args.verbose:
        return "DEBUG"
    if args.quiet:
        return "WARNING"
    return "INFO"


def get_terminal_size() -> int:
    try:
        n_cols = os.get_terminal_size().columns
    except OSError:
        n_cols = 80
    return n_cols


def configure_stdlib_base_logger(
    level: LevelType = "INFO",
    format="%(levelname)s: %(message)s",  # noqa  # pylint: disable=redefined-builtin
) -> None:
    """
    Configure the standard library logger.

    Args:
        level: minimum level to log
        format: message formatting
            (see https://docs.python.org/3/library/logging.html#logrecord-attributes)
    """
    level = get_level_as_int(level)
    logging.basicConfig(level=level, format=format)


def silence_stdlib_loggers(
    search_str: str,
    regex_mode: bool = False,
    level: LevelType = logging.ERROR,
    verbose: bool = False,
):
    # the packg.log module should be usable anywhere so we cannot import other packg modules here
    # otherwise we would cause a circular import. so we put the import here inside the function.
    from packg.iotools.pathspec_matcher import make_pathspec

    spec: PathSpec = make_pathspec([search_str], regex_mode=regex_mode)
    # print([p.regex for p in spec.patterns])
    level = get_level_as_int(level)
    for name in logging.root.manager.loggerDict.keys():  # pylint: disable=no-member
        loggr = logging.getLogger(name)
        if spec.match_file(name):
            loggr.setLevel(level)
            if verbose:
                print(f"Set verbosity to {level} for logger '{name}'")


# the default loguru logger level is debug, but it should be error.
# use configure_logger function anywhere to change the logger level.
configure_logger(level=logging.ERROR, sink=sys.stderr, format=SHORTEST_FORMAT, colorize=True)
logger = loguru_logger
logger.warn = deprecated("Use logger.warning instead")(logger.warning)



from attrs import define
from loguru import logger

from packg.iotools import loads_json
from packg.log import SHORTEST_FORMAT, configure_logger, get_logger_level_from_args
from packg.system import systemcall_with_assert
from typedparser import VerboseQuietArgs, TypedParser


@define
class Args(VerboseQuietArgs):
    pass


def main():
    parser = TypedParser.create_parser(Args, description=__doc__)
    args: Args = parser.parse_args()
    configure_logger(level=get_logger_level_from_args(args), format=SHORTEST_FORMAT)
    logger.info(f"{args}")

    jsondata, _, _ = systemcall_with_assert("pip list --format json")
    data = loads_json(jsondata)
    names = []
    for pkg in data:  # e.g. {"name": "a", "version": "3.1.2", " "editable_project_location": "..."}
        if "editable_project_location" in pkg:
            logger.warning(f"SKIP {pkg}")
            continue
        name = pkg["name"]
        names.append(name)
    final_cmd = f"pip install -U {' '.join(names)}"
    print()
    print(final_cmd)
    print()
    for name in names:
        print(f"pip install -U {name} --no-deps")
    print()


if __name__ == "__main__":
    main()



from __future__ import annotations


def dict_to_str_comma_equals(in_dict: dict[str, any] | list[tuple[any, any]]):
    try:
        items = in_dict.items()
    except AttributeError:
        items = in_dict
    str_list = []
    for k, v in items:
        str_list.append(f"{k}={v}")
    return ", ".join(str_list)



import shutil
from packg.log import logger


def format_pseudo_table(items, max_width=None, padding=2, indent=0):
    if len(items) == 0:
        logger.warning(f"(Empty table supplied to format_pseudo_table)")
        return ""
    if isinstance(items, str):
        raise ValueError(f"Expected list of strings, got string: {items}")
    if max_width is None:
        max_width = shutil.get_terminal_size().columns
    max_item_length = max(len(item) for item in items)
    num_columns = max_width // (max_item_length + padding)
    num_rows = (len(items) + num_columns - 1) // num_columns
    # Print items in a pseudo-table format
    lines = []
    for row in range(num_rows):
        row_items = [
            items[row + num_rows * col]
            for col in range(num_columns)
            if row + num_rows * col < len(items)
        ]
        lines.append(
            " " * indent + "".join(f"{item:<{max_item_length + padding}}" for item in row_items)
        )
    return "\n".join(lines)


def main():
    # todo turn into test file
    # todo test witg empty list
    items = [
        "file1.txt",
        "file2.jpg",
        "file3.pdf",
        "file4.do4234cx",
        "file5.mp3",
        "file6.png",
        "file7.zip",
        "file8.tsdfsar.gz",
        "file9.mov",
        "file10.avi",
        "file11.mkv",
        "file12.234txt",
        "file13.jpg",
        "file14.pdf",
        "file15.docx",
        "file16.mp3",
        "file17.png",
        "filsfde18.zip",
        "fi234234le19.tar.gz",
        "filesdfasdasdf20.mov",
    ]
    here_output = format_pseudo_table(items)
    print(here_output)
    gt_output = """\
file1.txt             file8.tsdfsar.gz      file15.docx           
file2.jpg             file9.mov             file16.mp3            
file3.pdf             file10.avi            file17.png            
file4.do4234cx        file11.mkv            filsfde18.zip         
file5.mp3             file12.234txt         fi234234le19.tar.gz   
file6.png             file13.jpg            filesdfasdasdf20.mov  
file7.zip             file14.pdf       """
    assert here_output.strip() == gt_output.strip()


if __name__ == "__main__":
    main()



"""
Utilities for python packages.

todo split this into multiple, make the get_module... thing faster and depend on less.
thenr eenable the logger.
"""

from __future__ import annotations

import importlib
import os
import re
import subprocess
import sys
import time
from pathlib import Path
from typing import Optional

import requests

from packg.iotools import sort_file_paths_with_dirs_separated
from packg.misc import format_exception
from packg.strings import create_nested_abbreviations
from packg.testing.import_from_source import recurse_modules


def run_package(main_file, run_dir="cli", recursive=True, run_dir_only=True, abbreviations=True):
    """
    todo remove this. it is too slow and ineffective. solve the problem in bash,
         or use the way simpler approach to import the main function and run it.

    Create a command line interface for a package given a directory of scripts.

    Check the run directory, list scripts, create shortcuts, run if given shortcut.

    Args:
        main_file: __file__ attribute of __main__.py e.g.
            /path/to/lmbtools/python/lmbtools/__main__.py
        run_dir: submodule with run scripts
        recursive: if True, recursively search for scripts
        run_dir_only: if True, only enable running scripts in the run_dir
        abbreviations: if True, create abbreviations for the scripts
    """
    main_file = Path(main_file)
    print(f"main_file: {main_file}")
    package_dir = main_file.parent
    package_name = package_dir.name
    package_scripts_dir = package_dir / run_dir
    glob_str = "**/*.py" if recursive else "*.py"

    package_scripts = [
        f.relative_to(package_scripts_dir)
        for f in package_scripts_dir.glob(glob_str)
        if not f.name.startswith("__")
    ]
    sorted_scripts = sort_file_paths_with_dirs_separated(package_scripts, dirs_first=False)
    package_scripts = [f.as_posix()[:-3].replace("/", ".") for f in sorted_scripts]

    abbrevs = None
    if abbreviations:
        abbrevs = create_nested_abbreviations(package_scripts)

    args = sys.argv[1:]
    if len(args) > 0:
        abbrev_arg = args[0]
        if abbrevs is not None and abbrev_arg in abbrevs:
            script = abbrevs[abbrev_arg]
            target = f"{package_name}.{run_dir}.{script}"
            run_script(target, args)
            return

        target = f"{package_name}.{abbrev_arg}"
        if run_dir_only:
            target = f"{package_name}.{run_dir}.{abbrev_arg}"
        try:
            importlib.import_module(target)
            found = True
        except ModuleNotFoundError as e:
            # must figure out whether the error is due to the module not existing
            # or due to some import failing.
            if f"'{target}'" in str(e):
                # No module named (the module we are trying to run) -> module not found
                found = False
            else:
                # No module named (some other module) -> module found, but has import error
                # run it to trigger the error
                found = True
        if found:
            run_script(target, args)
            return
        print(f"Error, script not found: {abbrev_arg} and {target}")

    # no arguments, error, show help
    if abbreviations:
        print(f"Usage option 1: {package_name} shortcut [args]")
        print(f"Usage option 2: {package_name} path.to.module [args]")
        print()
        print(f"shortcut   script")
        for abbrev, f in abbrevs.items():
            if run_dir_only:
                print(f"{abbrev:10s} {f}")
            else:
                print(f"{abbrev:10s} {run_dir}.{f}")
        return

    # show help but without abbreviations
    print(f"Usage: {package_name} path.to.module [args]")
    print()
    print(f"Available scripts:")
    for f in package_scripts:
        if run_dir_only:
            print(f"{f}")
        else:
            print(f"{run_dir}.{f}")


def run_script(target, args):
    cmd = ["python", "-m", target, *args[1:]]
    print(f"$ {' '.join(cmd)}")

    subprocess.run(cmd)

    # # run by importing main() function
    # print(f"Running {target}")
    # sys.argv = [sys.argv[0]] + args[1:]
    # module = importlib.import_module(target)
    # main = getattr(module, "main")
    # main()


def _recurse_modules_remove_root_package(package: str, packages_only: bool = False):
    for candidate in recurse_modules(package, packages_only=packages_only):
        if candidate == package:
            continue
        candidate = candidate.removeprefix(f"{package}.")
        if candidate == "__main__":
            continue
        yield candidate


def get_modules_for_autocomplete(package: str, run_dir: Optional[str] = None):
    packages_only = set(_recurse_modules_remove_root_package(package, packages_only=True))
    all_modules = set(_recurse_modules_remove_root_package(package))
    output_modules = []

    for m in sorted(all_modules):
        if m in packages_only:
            # mlog = f"{package}.{m}"
            if f"{m}.__main__" not in all_modules:
                # print(f"SKIP package: {mlog}")
                continue
            # print(f"ADD package (has __main__): {mlog}")
        if m.endswith("__main__"):
            # print(f"SKIP __main__ file itself: {mlog}")
            continue
        output_modules.append(m)
    output_modules.sort()

    if run_dir is not None:
        new_output_modules = []
        for output_module in output_modules:
            if output_module.startswith(run_dir):
                new_output_modules.append(output_module)
        output_modules = new_output_modules
    return output_modules


def create_bash_autocomplete_script(
    package: str, function_name: str = None, command_name: str = None, run_dir: Optional[str] = None
) -> str:
    """
    Create a bash script for autocompletion of a python package.

    Workflow where this is useful:
    - src/packg/__main__.py calls run_package, so `python -m packg run.script` has the same effect
    as `python -m packg.run.script`
    - pyproject.toml defines script [project.scripts] packg = "packg.__main__:main"
    so now `packg run.script` has the same effect.
    - this function now can be used to create src/packg/autocomplete.sh which, when sourced,
    provides autocompletion for the `packg` command.

    Args:
        package: package to create the autocomplete for
        function_name: default _{package}
        command_name: default {package}
        run_dir: only create autocompletion for this directory

    Returns:

    """
    output_modules = get_modules_for_autocomplete(package, run_dir=run_dir)
    if function_name is None:
        function_name = f"_{package}"
    if command_name is None:
        command_name = package
    ob, cb = "{", "}"
    autocomplete_script = f"""
# source this file to enable autocompletion for {package} command
{function_name}() {ob}
    local cur prev opts
    _init_completion || return
    # complete first argument with script
    if [ $COMP_CWORD -eq 1 ]; then
        opts="{' '.join(output_modules)}"
        COMPREPLY=( $( compgen -W "${ob}opts{cb}" -- "${ob}cur{cb}") )
        return 0
    fi
    # # otherwise complete with filesystem
    # COMPREPLY=( $(compgen -f -- "${ob}cur{cb}") )
    _filedir
    return 0
{cb}

complete -F {function_name} {command_name}
"""
    return autocomplete_script


def create_new_bash_autocomplete_script(
    packages: list[str],
    command_name: str = "py",
    function_name: Optional[str] = None,
    run_dir: Optional[str] = None,
) -> str:
    """
    New version: allinone e.g. alias py and then complete for all installed packages.

    Args:
        packages: packages to create the autocomplete for
        function_name: default _{package}
        command_name: default {package}
        run_dir: only create autocompletion for this directory

    Returns:

    """
    all_output_modules = []
    for package in packages:
        output_modules = get_modules_for_autocomplete(package, run_dir=run_dir)
        output_modules = [f"{package}.{m}" for m in output_modules]
        # logger.debug(f"package={package} output_modules={output_modules}")
        if len(output_modules) == 0:
            print(f"WARN: Nothing found for package {package} in {run_dir}, is it installed?")
        all_output_modules.extend(output_modules)
    if function_name is None:
        function_name = f"_{command_name}"

    ob, cb = "{", "}"
    autocomplete_script = f"""
{function_name}() {ob}
    local cur prev opts
    _init_completion || return
    # complete second argument with script, iff first argument is -m
    if [[ $COMP_CWORD -eq 2 && $prev == "-m" ]]; then
        opts="{' '.join(all_output_modules)}"
        COMPREPLY=( $( compgen -W "${ob}opts{cb}" -- "${ob}cur{cb}") )
        return 0
    fi
    # # otherwise complete with filesystem
    # COMPREPLY=( $(compgen -f -- "${ob}cur{cb}") )
    _filedir
    return 0
{cb}

complete -F {function_name} {command_name}
"""
    return autocomplete_script


FILEDIR_AUTOCOMPLETE = r"""
_filedir()  # source: ubuntu 2004 /usr/share/bash-completion/bash_completion
{
    local IFS=$'\n'
    _tilde "$cur" || return
    local -a toks
    local reset
    if [[ "$1" == -d ]]; then
        reset=$(shopt -po noglob); set -o noglob
        toks=( $(compgen -d -- "$cur") )
        IFS=' '; $reset; IFS=$'\n'
    else
        local quoted
        _quote_readline_by_ref "$cur" quoted
        # Munge xspec to contain uppercase version too
        # http://thread.gmane.org/gmane.comp.shells.bash.bugs/15294/focus=15306
        local xspec=${1:+"!*.@($1|${1^^})"} plusdirs=()
        # Use plusdirs to get dir completions if we have a xspec; if we don't,
        # there's no need, dirs come along with other completions. Don't use
        # plusdirs quite yet if fallback is in use though, in order to not ruin
        # the fallback condition with the "plus" dirs.
        local opts=( -f -X "$xspec" )
        [[ $xspec ]] && plusdirs=(-o plusdirs)
        [[ ${COMP_FILEDIR_FALLBACK-} ]] || opts+=( "${plusdirs[@]}" )
        reset=$(shopt -po noglob); set -o noglob
        toks+=( $(compgen "${opts[@]}" -- $quoted) )
        IFS=' '; $reset; IFS=$'\n'
        # Try without filter if it failed to produce anything and configured to
        [[ -n ${COMP_FILEDIR_FALLBACK:-} && -n "$1" && ${#toks[@]} -lt 1 ]] && {
            reset=$(shopt -po noglob); set -o noglob
            toks+=( $(compgen -f "${plusdirs[@]}" -- $quoted) )
            IFS=' '; $reset; IFS=$'\n'
        }
    fi
    if [[ ${#toks[@]} -ne 0 ]]; then
        # 2>/dev/null for direct invocation, e.g. in the _filedir unit test
        compopt -o filenames 2>/dev/null
        COMPREPLY+=( "${toks[@]}" )
    fi
} # _filedir()

"""


def _get_raw_shields_io_output(package: str):
    # todo wrap this request code into a separate function in packg.web
    url = f"https://img.shields.io/pypi/v/{package}"
    print(f"Finding pypi version via {url}")
    counter = 0
    while True:
        try:
            response = requests.get(url, timeout=10)
            break
        except Exception as e:
            print(f"ERROR requesting {url}: {format_exception(e)}")
        counter += 1
        time.sleep(1)
    return response.text


_RE_TEXT = re.compile(r"<text.*?>.*?</text>")
_RE_TEXT_VERSION = re.compile(r"<text.*?>v([0-9]+.[0-9]+.[0-9]+)</text>")


def find_pypi_package_version(package: str) -> Optional[str]:
    """
    Find the latest version of a package on pypi.

    pip search was removed. An alternative is to query the shields.io service and parse the
    output svg.
    https://img.shields.io/pypi/v/numpy

    Args:
        package: package name

    Returns:
        version string

    """
    raw = _get_raw_shields_io_output(package)
    # find all occurrences of <text>...</text>
    matches = _RE_TEXT.findall(raw)
    # the last one is the displayed version
    version = matches[-1]
    # find the version number
    match = _RE_TEXT_VERSION.search(version)
    if match:
        version = match.group(1)
        return version
    print(f"ERROR: Could not find version for {package} on shields.io")
    return None


def find_top_level_package(file_path, verbose: bool = True):
    """
    Args:
        file_path: __file__ of the module or script you want to find the top-level package for
    Returns:
    """
    print_fn = print if verbose else lambda *args, **kwargs: None
    current_file_path = os.path.abspath(file_path)
    print_fn(f"Current file path: {current_file_path}")
    package = str(__package__)
    print_fn(f"__package__: {package}")
    name = __name__
    print_fn(f"__name__: {name}")

    # Attempt to determine the top-level package
    if package:
        top_level_package = package.split(".", maxsplit=1)[0]
    else:
        # If __package__ is None, infer top-level package from the file structure
        top_level_package = os.path.basename(os.path.dirname(current_file_path))

    print_fn(f"Top-level package: {top_level_package}")
    return top_level_package


def main():
    print(_get_raw_shields_io_output("numpy"))


if __name__ == "__main__":
    main()



"""
import all installed packages (useful to check for errors, to pre-generate cache etc.)

Example usage:
    for i in $(cat import_all.txt); do echo $i; python -c "import $i; print($i.__name__)"; done
"""

import tempfile
from importlib.metadata import entry_points

from attrs import define
from loguru import logger

from packg.iotools import make_git_pathspec
from packg.log import SHORTEST_FORMAT, configure_logger, get_logger_level_from_args
from packg.system import systemcall
from packg.testing import recurse_modules
from packg.tqdmext import tqdm_max_ncols
from typedparser import VerboseQuietArgs, add_argument, TypedParser


@define
class Args(VerboseQuietArgs):
    filter_imports: str = add_argument(shortcut="-f", type=str, help="Filter imports", default="")
    filter_package_list: str = add_argument(
        shortcut="-p", type=str, help="Filter package list", default=""
    )
    recursive: bool = add_argument(
        shortcut="-r", action="store_true", help="Recursively import all"
    )


def main():
    parser = TypedParser.create_parser(Args, description=__doc__)
    args: Args = parser.parse_args()
    configure_logger(level=get_logger_level_from_args(args), format=SHORTEST_FORMAT)
    logger.info(f"{args}")

    eps = entry_points()
    all_vs = []
    for _k, vs in eps.items():
        for v in vs:
            all_vs.append(v.value.split(":")[0])
    all_vs = sorted(set(all_vs))
    more_vs = []
    spec = None
    if len(args.filter_imports) > 0:
        filter_list = args.filter_imports.split(",")
        spec = make_git_pathspec(filter_list)
    filter_package_set = set()
    if len(args.filter_package_list) > 0:
        filter_package_set = set(args.filter_package_list.split(","))
    for v in all_vs:
        current = []
        for part in v.split("."):
            current.append(part)
            to_add = ".".join(current)
            if spec is not None and not spec.match_file(to_add):
                continue
            if len(filter_package_set) > 0 and to_add not in filter_package_set:
                continue
            more_vs.append(to_add)
    more_vs = sorted(set(more_vs))
    with tempfile.NamedTemporaryFile(
        prefix="python_import_all_file", suffix=".txt", delete=True, mode="wt", encoding="utf-8"
    ) as fh:
        print(f"Wrote: {fh.name}")
        fh.write("\n".join(more_vs))
        fh.flush()

    all_module_dict = {}
    for v in more_vs:
        module_list = [v]
        if args.recursive:
            module_list += list(recurse_modules(v, ignore_tests=True, packages_only=False))
        all_module_dict[v] = module_list
    total_len = sum(len(v) for v in all_module_dict.values())

    pbar = tqdm_max_ncols(desc="Importing", total=total_len)
    for v, module_list in all_module_dict.items():
        pbar.write(f"---------- importing {v} len {len(module_list)} total {total_len} ----------")
        strs = []
        for mod in module_list:
            strs += [
                "try:",
                f"    import {mod}",
                f'    print("{mod}", end=" ", flush=True)',
                "except Exception as e:",
                f"    from packg import format_exception",
                f'    print(f"Error importing {mod}")',
                f"    print(format_exception(e))",
                f"    print()",
                f"",
            ]
        final_str = "\n".join(strs)
        out, err, retcode = systemcall(f"python -c '{final_str}'")
        outputs = []
        if retcode != 0:
            outputs.append(f"ERROR CODE: {retcode}")
        if out.strip() != "":
            outputs.append(out.strip())
        if err.strip() != "":
            outputs.append(err.strip())
        if len(outputs) > 0:
            pbar.write("\n".join(outputs))
        pbar.update(len(module_list))


if __name__ == "__main__":
    main()



"""
Count lines per file endings in a directory.
"""
from __future__ import annotations

from collections import defaultdict
from pathlib import Path
from typing import Optional

from attrs import define
from loguru import logger

from packg import format_exception
from packg.iotools import make_index
from packg.iotools.encoding import (
    detect_encoding_and_read_file,
    detect_encoding_if_needed_and_read_file,
)
from packg.log import SHORTEST_FORMAT, configure_logger, get_logger_level_from_args
from packg.tqdmext import tqdm_max_ncols
from typedparser import VerboseQuietArgs, add_argument, TypedParser


@define
class Args(VerboseQuietArgs):
    folder: Path = add_argument("folder", type=str, help="Directory to check")
    ignore_case: bool = add_argument(shortcut="-i", action="store_true", help="Ignore case")
    endings: Optional[list[str]] = add_argument(
        shortcut="-e", action="append", help="count files with these endings"
    )
    write_fix: bool = add_argument(
        shortcut="-w", action="store_true", help="Write files with fixed encoding"
    )
    check_all: bool = add_argument(
        shortcut="-a", action="store_true", help="Check all files for encoding"
    )


def main():
    parser = TypedParser.create_parser(Args, description=__doc__)
    args: Args = parser.parse_args()
    configure_logger(level=get_logger_level_from_args(args), format=SHORTEST_FORMAT)
    logger.info(f"{args}")

    if args.endings is None or len(args.endings) == 0:
        raise ValueError("No endings provided, use e.g. -e py -e ipynb")
    endings_set = set(args.endings)

    folder = Path(args.folder)
    if not folder.is_dir():
        raise ValueError(f"Not a dir: {folder}")

    file_index = make_index(folder)
    print()
    ending2filenames = defaultdict(list)
    for filename in file_index.keys():
        splits = Path(filename).name.split(".")
        if len(splits) == 1:
            ending = ""
        else:
            ending = splits[-1]
        if ending not in endings_set:
            continue
        if args.ignore_case:
            ending = ending.lower()
        ending2filenames[ending].append(filename)
    ending2filenames = dict(ending2filenames)
    for ending, filenames in ending2filenames.items():
        num_lines_total = 0
        for filename in tqdm_max_ncols(filenames, desc=ending):
            full_file = folder / filename
            try:
                if args.check_all:
                    content, encoding = detect_encoding_and_read_file(full_file)
                else:
                    content, encoding = detect_encoding_if_needed_and_read_file(full_file)
            except ValueError as e:
                logger.error(f"Failed reading file: {format_exception(e)}")
                continue

            content_unix = content.replace("\r\n", "\n")
            num_lines = len(content_unix.splitlines())
            num_lines_total += num_lines
            if encoding in set(("utf-8", "ascii", None)) and content == content_unix:
                # file is already fine
                continue
            # file should be fixed
            win_lineendings = content.count("\r\n")
            msg = f"encoding={encoding} win_lineendings={win_lineendings} full_file={full_file}"
            if not args.write_fix:
                logger.warning(f"{msg} - would fix but no -w/--write_fix")
                continue
            logger.warning(f"{msg} - fixing")
            full_file.write_text(content_unix, encoding="utf-8")

        print(f"{ending}: {len(filenames)} files, {num_lines_total:_d} lines")


if __name__ == "__main__":
    main()



"""
Create a bash autocompletion script.

Given the files __main__.py and pyproject.toml, after installation,
the command `packg my.module` will run `python -m packg.my.module`.
This script will create the autocompletion script over all modules in the package.
"""

import importlib
from pathlib import Path
from typing import Optional

from attrs import define
from loguru import logger

from packg.log import SHORTEST_FORMAT, configure_logger, get_logger_level_from_args
from packg.packaging import FILEDIR_AUTOCOMPLETE, create_bash_autocomplete_script
from typedparser import VerboseQuietArgs, add_argument, TypedParser


@define
class Args(VerboseQuietArgs):
    packages: str = add_argument(shortcut="-p", type=str, default="packg,visiontext")
    run_dir: Optional[str] = add_argument(
        shortcut="-r", type=str, help="Only create autocompletion for this directory"
    )


def main():
    parser = TypedParser.create_parser(Args, description=__doc__)
    args: Args = parser.parse_args()
    configure_logger(level=get_logger_level_from_args(args), format=SHORTEST_FORMAT)
    logger.info(f"{args}")

    packages = args.packages.split(",")
    for package in packages:
        mod = importlib.import_module(package)
        # print(mod.__file__)
        autocomplete_file = Path(mod.__file__).parent.parent.parent / f"autocomplete_{package}.sh"
        autocomplete_script = create_bash_autocomplete_script(
            package, command_name=package, run_dir=args.run_dir
        )
        Path(autocomplete_file).write_text(
            "\n".join([autocomplete_script, FILEDIR_AUTOCOMPLETE]), encoding="utf-8"
        )
        logger.info(f"Created: {autocomplete_file}")
    # logger.debug(autocomplete_script)


if __name__ == "__main__":
    main()



from __future__ import annotations

from pathlib import Path

import chardet


def detect_encoding(file_path):
    with open(file_path, "rb") as f:
        result = chardet.detect(f.read())
        return result["encoding"]


def detect_encoding_and_read_file(file: Path) -> tuple[str, str]:
    encoding = detect_encoding(file)
    try:
        content = file.read_text(encoding=encoding)
    except UnicodeDecodeError as e:
        raise ValueError(f"Finally failed reading {file} with encoding {encoding}") from e
    return content, encoding


def detect_encoding_if_needed_and_read_file(file: Path) -> str:
    try:
        content = file.read_text(encoding="utf-8")
        encoding = "utf-8"
        return content, encoding
    except UnicodeDecodeError:
        return detect_encoding_and_read_file(file)



from __future__ import annotations

import csv
import io
from csv import Error as CsvError


def format_to_tsv(data: list[list[any]]) -> str:
    """
    Args:
        data: 2d list of list that should be formatted to tsv

    Returns:
        tsv formatted string
    """
    output = io.StringIO()
    writer = csv.writer(output, delimiter="\t", quoting=csv.QUOTE_MINIMAL, lineterminator="\n")
    for i, row in enumerate(data):
        if isinstance(row, str):
            raise CsvError(f"Expected list, got str for row {i} content {row}")
        writer.writerow(row)
    formatted_tsv = output.getvalue()
    output.close()
    return formatted_tsv



"""
Shortcut to create a pathspec from a list of patterns with either .gitignore or regex syntax.

https://pypi.org/project/pathspec/#description

"""

from __future__ import annotations

from pathlib import Path
from typing import Iterable, Union, List, Optional, Tuple, Dict, Any

from attr import define
from pathspec import PathSpec, patterns as p_patterns, RegexPattern, Pattern

from packg.typext import PathType
from typedparser import add_argument


class PathSpecRepr(PathSpec):
    def __str__(self):
        return repr_pathspec(self)

    def __repr__(self):
        return repr_pathspec(self)


def repr_pathspec(pspec: PathSpec):
    patts: list[Pattern] = pspec.patterns
    reps = []
    for patt in patts:
        patt_rep = f"Unknown"
        if hasattr(patt, "pattern"):
            patt_rep = str(patt.pattern)
        reps.append(f"{type(patt).__name__}({patt_rep})")
    return f"PathSpec({', '.join(reps)})"


def make_git_pathspec(patterns: List[str]) -> PathSpecRepr:
    """
    >>> spec1: PathSpecRepr = make_git_pathspec(["hello*world"])
    >>> print(spec1.match_file("hello_world"))
    True
    >>> print(spec1.match_file("hello_world.txt"))
    False

    Args:
        patterns:

    Returns:

    """
    spec = PathSpecRepr.from_lines(p_patterns.GitWildMatchPattern, patterns)
    return spec


def make_regex_pathspec(patterns: List[str]) -> PathSpecRepr:
    spec = PathSpecRepr.from_lines(RegexPattern, patterns)
    return spec


def make_pathspec(patterns: List[str], regex_mode: bool = False) -> PathSpecRepr:
    """
    Create PathSpec object that can be used to match or search files.

    Args:
        patterns: separated list of patterns to search for
        regex_mode: False (default) = .gitignore syntax, True = regex

    Returns:
        PathSpec object
    """
    if regex_mode:
        return make_regex_pathspec(patterns)
    return make_git_pathspec(patterns)


def make_and_apply_pathspecs(
    paths: List[PathType],
    include_git: Optional[List[str]] = None,
    include_regex: Optional[List[str]] = None,
    exclude_git: Optional[List[str]] = None,
    exclude_regex: Optional[List[str]] = None,
    exclude_gitignore_file: Optional[PathType] = None,
) -> Iterable[Path]:
    specs = make_pathspecs(
        include_git=include_git,
        include_regex=include_regex,
        exclude_git=exclude_git,
        exclude_regex=exclude_regex,
        exclude_gitignore_file=exclude_gitignore_file,
    )
    paths = apply_pathspecs(paths, specs)
    return paths


SPECLISTTYPE = List[Tuple[PathSpec, bool]]


def make_pathspecs(
    include_git: Optional[List[str]] = None,
    include_regex: Optional[List[str]] = None,
    exclude_git: Optional[List[str]] = None,
    exclude_regex: Optional[List[str]] = None,
    exclude_gitignore_file: Optional[PathType] = None,
) -> SPECLISTTYPE:
    """
    Builds PathSpecs based on inputs to include or exclude.

    Returns:
        List of tuples with PathSpec and negate flag

    """
    specs = []
    if include_git is not None and len(include_git) > 0:
        spec = make_git_pathspec(include_git)
        specs.append((spec, False))
    if include_regex is not None and len(include_regex) > 0:
        spec = make_regex_pathspec(include_regex)
        specs.append((spec, False))
    if exclude_git is not None and len(exclude_git) > 0:
        spec = make_git_pathspec(exclude_git)
        specs.append((spec, True))
    if exclude_regex is not None and len(exclude_regex) > 0:
        spec = make_regex_pathspec(exclude_regex)
        specs.append((spec, True))
    if exclude_gitignore_file is not None:
        spec = make_git_pathspec(exclude_gitignore_file.read_text(encoding="utf-8").splitlines())
        specs.append((spec, True))
    return specs


def apply_pathspecs(paths: List[PathType], specs: SPECLISTTYPE) -> Iterable[Path]:
    for spec, negate in specs:
        paths = spec.match_files(paths, negate=negate)
    return paths


@define(slots=False)
class PathSpecArgs:
    exclude_git: List[str] = add_argument(
        shortcut="-x", default=[], action="append", help="Git-like pathspec list to exclude files."
    )
    exclude_regex: List[str] = add_argument(
        shortcut="-X", default=[], action="append", help="Regex pathspec list to exclude files."
    )
    include_git: List[str] = add_argument(
        shortcut="-i", default=[], action="append", help="Git-like pathspec list to include files."
    )
    include_regex: List[str] = add_argument(
        shortcut="-I", default=[], action="append", help="Regex pathspec list to include files."
    )
    exclude_gitignore_file: Optional[Path] = add_argument(
        shortcut="-g", type=str, default=None, help="Gitignore file for matching files"
    )


def expand_pathspec_args(args: PathSpecArgs) -> Dict[str, Any]:
    return dict(
        exclude_git=args.exclude_git,
        exclude_regex=args.exclude_regex,
        include_git=args.include_git,
        include_regex=args.include_regex,
        exclude_gitignore_file=args.exclude_gitignore_file,
    )


class PathSpecWithConversion:
    """
    Wrapper for pathspec.PathSpec with argument conversion to and from pathlib.Path

    Usage:
        matches = spec.match_tree("path/to/dir")
        matches = spec.match_files(file_paths)
        is_matched = spec.match_file("path/to/file")

    Args:
        patterns: separated list of patterns to search for
        regex_mode: False (default) = .gitignore syntax, True = regex
        output_mode: How to return the matched files. One of "str", "path", "str_posix"
    """

    CONVERTERS = {
        "str": str,
        "path": Path,
        "str_posix": lambda input_path: Path(input_path).as_posix(),
    }

    def __init__(
        self,
        patterns: List[str],
        regex_mode: bool = False,
        input_mode: str = "str",
        output_mode: str = "path",
    ):
        self.patterns = patterns
        self.spec = make_pathspec(patterns, regex_mode=regex_mode)
        self.converter_in = self.CONVERTERS[input_mode]
        self.converter_out = self.CONVERTERS[output_mode]

    def match_tree(self, path: Union[str, Path]) -> Iterable[Path]:
        """Matches all files in path recursively, automatically excludes folders."""
        path = self.converter_in(path)
        for file in self.spec.match_tree(path):
            yield self.converter_out(file)

    def match_files(self, paths: List[Union[str, Path]]) -> Iterable[Path]:
        """Yields all entries in paths that match, does not exclude folders."""
        paths = [self.converter_in(path) for path in paths]
        for file in self.spec.match_files(paths):
            yield self.converter_out(file)

    def match_file(self, path: Union[str, Path]) -> bool:
        """Matches this single path, does not exclude folders."""
        return self.spec.match_file(self.converter_in(path))
